@@ -1,4 +1,5 @@
 import io
+import os
 import logging
 from zipfile import ZipFile, ZIP_STORED, ZIP_LZMA
 import json
@@ -6,15 +7,120 @@ import tempfile
 import sys
 import subprocess
 
+from flex.flex import FlexFile
+from . import __version__
 
 import numpy as np
 
 logger = logging.getLogger(__name__)
 
+
+def to_flex(sme):
+    header = {}
+    extensions = {}
+
+    for name in sme._names:
+        value = sme[name]
+        if isinstance(value, IPersist):
+            extensions[name] = value._save()
+        elif value is not None:
+            header[name] = value
+
+    ff = FlexFile(header, extensions)
+    return ff
+
+
+def from_flex(ff, sme):
+    header = ff.header
+    extensions = ff.extensions
+    for name in sme._names:
+        if name in updates.keys():
+            name = updates[name]
+        if name in header.keys():
+            sme[name] = header[name]
+        elif name in extensions.keys():
+            if sme[name] is not None:
+                sme[name] = sme[name]._load(extensions[name])
+            else:
+                sme[name] = extensions[name]
+    return sme
+
+
+def save(fname, sme, compressed=False):
+    """
+    Create a folder structure inside a tarfile
+    See flex-format for details
+
+    Parameters
+    ----------
+    filename : str
+        Filename of the final file
+    sme : SME_Structure
+        sme structure to save
+    compressed : bool, optional
+        whether to compress the output
+    """
+    ff = to_flex(sme)
+    ff.write(fname)
+
+
+def load(fname, sme):
+    """
+    Load the SME Structure from disk
+
+    Parameters
+    ----------
+    fname : str
+        file to load
+    sme : SME_Structure
+        empty sme structure with default values set
+
+    Returns
+    -------
+    sme :  SME_Structure
+        loaded sme structure
+    """
+    try:
+        ff = FlexFile.read(fname)
+        sme = from_flex(ff, sme)
+        ff.close()
+        return sme
+    except Exception as ex:
+        logger.error(ex)
+        try:
+            sme = load_v1(fname, sme)
+        except:
+            raise ex
+        return sme
+
+
 # Update this if the names in sme change
 updates = {"idlver": "system_info"}
 
 
+class IPersist:
+    def _save(self):
+        raise NotImplementedError
+
+    @classmethod
+    def _load(cls, ext):
+        raise NotImplementedError
+
+    def _save_v1(self, file, folder=""):
+        saves_v1(file, self, folder)
+
+    @classmethod
+    def _load_v1(cls, file, names, folder=""):
+        logger.setLevel(logging.INFO)
+        data = cls()  # TODO Suppress warnings
+        data = loads_v1(file, data, names, folder)
+        logger.setLevel(logging.NOTSET)
+        return data
+
+
+# Version 1 IO (Deprecated)
+
+
 def toBaseType(value):
     if value is None:
         return value
@@ -32,7 +138,7 @@ def toBaseType(value):
     return value
 
 
-def save(filename, data, folder="", compressed=False):
+def save_v1(filename, data, folder="", compressed=False):
     """
     Create a folder structure inside a zipfile
     Add .json and .npy and .npz files with the correct names
@@ -60,12 +166,12 @@ def save(filename, data, folder="", compressed=False):
         compression = ZIP_LZMA
 
     with ZipFile(filename, "w", compression) as file:
-        saves(file, data, folder=folder)
+        saves_v1(file, data, folder=folder)
 
 
 # TODO: this is specific for Collection type objects
 # Move this to Collection, and not here
-def saves(file, data, folder=""):
+def saves_v1(file, data, folder=""):
     if folder != "" and folder[-1] != "/":
         folder = folder + "/"
 
@@ -94,16 +200,16 @@ def saves(file, data, folder=""):
 
     for key, value in others.items():
         if value is not None:
-            value._save(file, f"{folder}{key}")
+            value._save_v1(file, f"{folder}{key}")
 
 
-def load(filename, data):
+def load_v1(filename, data):
     with ZipFile(filename, "r") as file:
         names = file.namelist()
-        return loads(file, data, names)
+        return loads_v1(file, data, names)
 
 
-def loads(file, data, names=None, folder=""):
+def loads_v1(file, data, names=None, folder=""):
     if folder != "" and folder[-1] != "/":
         folder = folder + "/"
     if names is None:
@@ -142,11 +248,14 @@ def loads(file, data, names=None, folder=""):
 
     for key, value in subdirs.items():
         data_key = updates.get(key, key)
-        data[data_key] = data[data_key]._load(file, value, folder=folder + key)
+        data[data_key] = data[data_key]._load_v1(file, value, folder=folder + key)
 
     return data
 
 
+# IDL IO
+
+
 def get_typecode(dtype):
     """ Get the IDL typecode for a given dtype """
     if dtype.name[:5] == "bytes":
@@ -214,8 +323,7 @@ def write_as_idl(sme):
     data arrays are stored in seperate temp files, and only the filename is passed to idl
     """
 
-    wind = np.cumsum(sme.wave.shape[1]) - 1
-    vrad_flag = {"none": -2, "whole": -1, "each": 0}[sme.vrad_flag]
+    vrad_flag = {"none": -2, "whole": -1, "each": 0, "fix": -2}[sme.vrad_flag]
     cscale_flag = {"none": -3, "fix": -3, "constant": 0, "linear": 1, "quadratic": 1,}[
         sme.cscale_flag
     ]
@@ -258,16 +366,10 @@ def write_as_idl(sme):
         "species": save_as_binary(sme.species),
         "atomic": save_as_binary(sme.atomic),
         "lande": save_as_binary(sme.linelist.lande),
-        "depth": save_as_binary(sme.linelist.depth),
         "lineref": save_as_binary(sme.linelist.reference),
         "short_line_format": {"short": 1, "long": 2}[sme.linelist.lineformat],
         "wran": sme.wran.tolist(),
         "mu": sme.mu.tolist() if sme.nmu > 1 else sme.mu[0],
-        "wave": save_as_binary(sme.wave.ravel()),
-        "wind": wind.tolist(),
-        "sob": save_as_binary(sme.spec.ravel()),
-        "uob": save_as_binary(sme.uncs.ravel()),
-        "mob": save_as_binary(sme.mask.ravel().astype("int16")),
         "obs_name": "",
         "obs_type": 0,
         "glob_free": fitvars,
@@ -280,15 +382,46 @@ def write_as_idl(sme):
         },
     }
 
+    if len(sme.nlte.elements) != 0:
+        idl_fields["nlte"] = {}
+
+        flags = np.zeros(99, dtype="int16")
+        grids = ["" for _ in range(99)]
+        for elem in sme.nlte.elements:
+            flags[sme.abund.elem_dict[elem]] = 1
+            grids[sme.abund.elem_dict[elem]] = sme.nlte.grids[elem]
+
+        idl_fields["nlte"]["nlte_elem_flags"] = save_as_binary(flags)
+        idl_fields["nlte"]["nlte_subgrid_size"] = save_as_binary(
+            sme.nlte.subgrid_size.astype("int16")
+        )
+        idl_fields["nlte"]["nlte_grids"] = grids
+        idl_fields["nlte"]["nlte_pro"] = "sme_nlte"
+
     # if sme.iptype is not None:
     #     idl_fields["iptype"] = sme.iptype
     #     idl_fields["ipres"] = sme.ipres
     #     # "ip_x": sme.ip_x,
     #     # "ip_y": sme.ip_y,
 
+    if sme.wave is not None:
+        wind = np.cumsum(sme.wave.shape[1]) - 1
+        idl_fields["wave"] = save_as_binary(sme.wave.ravel())
+        idl_fields["wind"] = wind.tolist()
+    if sme.spec is not None:
+        idl_fields["sob"] = save_as_binary(sme.spec.ravel())
+    if sme.uncs is not None:
+        idl_fields["uob"] = save_as_binary(sme.uncs.ravel())
+    if sme.mask is not None:
+        idl_fields["mob"] = save_as_binary(sme.mask.ravel().astype("int16"))
     if sme.synth is not None:
         idl_fields["smod"] = save_as_binary(sme.synth.ravel())
 
+    if "depth" in sme.linelist.columns:
+        idl_fields["depth"] = save_as_binary(sme.linelist.depth)
+    else:
+        idl_fields["depth"] = save_as_binary(np.ones(len(sme.linelist)))
+
     if sme.linelist.lineformat == "long":
         idl_fields.update(
             {
@@ -393,16 +526,3 @@ sme = new_sme\n"""
         subprocess.run(["idl", "-e", ".r %s" % tempname])
         # input("Wait for me...")
         clean_temps()
-
-
-class IPersist:
-    def _save(self, file, folder=""):
-        saves(file, self, folder)
-
-    @classmethod
-    def _load(cls, file, names, folder=""):
-        logger.setLevel(logging.INFO)
-        data = cls()  # TODO Suppress warnings
-        data = loads(file, data, names, folder)
-        logger.setLevel(logging.NOTSET)
-        return data
