@@ -13,6 +13,7 @@ from .abund import Abund, elements as abund_elem
 from .atmosphere.atmosphere import Atmosphere
 from .iliffe_vector import Iliffe_vector
 from .linelist.linelist import LineList
+from .nlte import NLTE
 
 from .data_structure import *
 
@@ -35,6 +36,8 @@ class Parameters(Collection):
     def __init__(self, **kwargs):
         monh = kwargs.pop("monh", kwargs.pop("feh", 0))
         abund = kwargs.pop("abund", "empty")
+        if "grav" in kwargs.keys() and "logg" not in kwargs.keys():
+            kwargs["logg"] = kwargs.pop("grav")
         super().__init__(**kwargs)
         self.abund = Abund(monh=monh, pattern=abund, type="sme")
 
@@ -67,109 +70,6 @@ class Parameters(Collection):
         return self.abund.citation()
 
 
-@CollectionFactory
-class NLTE(Collection):
-    # fmt: off
-    _fields = Collection._fields + [
-        ("elements", [], astype(list), this,
-            "list: elements for which nlte calculations will be performed"),
-        ("grids", {}, astype(dict), this,
-            "dict: nlte grid datafiles for each element"),
-        ("subgrid_size", [2, 2, 2, 2], array(4, int), this,
-            "array of shape (4,): defines size of nlte grid cache."
-            "Each entry is for one parameter abund, teff, logg, monh"),
-        ("flags", None, array(None, np.bool_), this,
-            "array: contains a flag for each line, whether it was calculated in NLTE (True) or not (False)")
-    ]
-    # fmt: on
-
-    _default_grids = {
-        "Al": "marcs2012_Al2017.grd",
-        "Fe": "marcs2012_Fe2016.grd",
-        "Li": "marcs2012_Li.grd",
-        "Mg": "marcs2012_Mg2016.grd",
-        "Na": "marcs2012p_t1.0_Na.grd",
-        "O": "marcs2012_O2015.grd",
-        "Ba": "marcs2012p_t1.0_Ba.grd",
-        "Ca": "marcs2012p_t1.0_Ca.grd",
-        "Si": "marcs2012_SI2016.grd",
-        "Ti": "marcs2012s_t2.0_Ti.grd",
-    }
-
-    def __init__(self, **kwargs):
-        super().__init__()
-
-        # Convert IDL keywords to Python
-        if "nlte_elem_flags" in kwargs.keys():
-            elements = kwargs["nlte_elem_flags"]
-            self.elements = [abund_elem[i] for i, j in enumerate(elements) if j == 1]
-
-        if "nlte_subgrid_size" in kwargs.keys():
-            self.subgrid_size = kwargs["nlte_subgrid_size"]
-
-        if "nlte_grids" in kwargs:
-            grids = kwargs["nlte_grids"]
-            if isinstance(grids, (list, np.ndarray)):
-                grids = {
-                    abund_elem[i]: name.decode()
-                    for i, name in enumerate(grids)
-                    if name != ""
-                }
-            self.grids = grids
-
-    def set_nlte(self, element, grid=None):
-        """
-        Add an element to the NLTE calculations
-
-        Parameters
-        ----------
-        element : str
-            The abbreviation of the element to add to the NLTE calculations
-        grid : str, optional
-            Filename of the NLTE data grid to use for this element
-            the file must be in nlte_grids directory
-            Defaults to a set of "known" files for some elements
-        """
-        if element in self.elements:
-            # Element already in NLTE
-            # Change grid if given
-            if grid is not None:
-                self.grids[element] = grid
-            return
-
-        if grid is None:
-            # Use default grid
-            if element not in NLTE._default_grids.keys():
-                raise ValueError(f"No default grid known for element {element}")
-            grid = NLTE._default_grids[element]
-            logger.info("Using default grid %s for element %s", grid, element)
-
-        if element not in self.elements:
-            self.elements += [element]
-        self.grids[element] = grid
-
-    def remove_nlte(self, element):
-        """
-        Remove an element from the NLTE calculations
-
-        Parameters
-        ----------
-        element : str
-            Abbreviation of the element to remove from NLTE
-        """
-        if element not in self.elements:
-            # Element not included in NLTE anyways
-            return
-
-        self.elements.remove(element)
-        self.grids.pop(element)
-
-    def citation(self, output="string"):
-        # TODO
-        citations = [self.grids[el] for el in self.elements]
-        return citations
-
-
 @CollectionFactory
 class Version(Collection):
     # fmt: off
@@ -204,26 +104,24 @@ class Version(Collection):
 class Fitresults(Collection):
     # fmt: off
     _fields = Collection._fields + [
-        ("maxiter", 100, astype(int), this, "int: maximum number of iterations in the solver"),
-        ("chisq", 0, asfloat, this, "float: reduced chi-square of the solution"),
-        ("uncertainties", None, array(None, float), this, "array of size(nfree,): uncertainties of the free parameters"),
+        ("maxiter", None, astype(int, allow_None=True), this, "int: maximum number of iterations in the solver"),
+        ("chisq", None, this, this, "float: reduced chi-square of the solution"),
+        ("parameters", None, this, this, "list: parameter names"),
+        ("values", None, array(None, float), this, "array: best fit values for the fit parameters"),
+        ("uncertainties", None, array(None, float), this, "array of size(nfree,): uncertainties of the free parameters bases on SME statistics"),
         ("covariance", None, array(None, float), this, "array of size (nfree, nfree): covariance matrix"),
         ("gradient", None, array(None, float), this, "array of size (nfree,): final gradients of the free parameters on the cost function"),
         ("derivative", None, array(None, float), this, "array of size (npoints, nfree): final Jacobian of each point and each parameter"),
         ("residuals", None, array(None, float), this, "array of size (npoints,): final residuals of the fit"),
-        ("punc2", None, this, this, "array: old school uncertainties")
+        ("fit_uncertainties", None, this, this, "array: uncertainties based solely on the least_squares fit")
     ]
     # fmt: on
 
     def clear(self):
         """ Reset all values to None """
-        self.maxiter = 100
-        self.chisq = 0
-        self.punc = None
-        self.covar = None
-        self.grad = None
-        self.pder = None
-        self.resid = None
+        for name in self._names:
+            default = [f[1] for f in self._fields if f[0] == name][0]
+            setattr(self, name, default)
 
 
 @CollectionFactory
@@ -231,9 +129,8 @@ class SME_Structure(Parameters):
     # fmt: off
     _fields = Parameters._fields + [
         ("id", dt.now(), asstr, this, "str: DateTime when this structure was created"),
-        ("object", "", asstr, this, "str: Name of the observed/simulated object"),
+        ("meta", {}, this, this, "dict: Arbitrary extra information"),
         ("version", __version__, this, this, "str: PySME version used to create this structure"),
-        ("vrad", 0, array(None, float), this, "array of size (nseg,): radial velocity of each segment in km/s"),
         ("vrad_flag", "none", lowercase(oneof(-2, -1, 0, "none", "each", "whole", "fix")), this,
             """str: flag that determines how the radial velocity is determined
 
@@ -242,12 +139,8 @@ class SME_Structure(Parameters):
                * "each": Determine radial velocity for each segment individually
                * "whole": Determine one radial velocity for the whole spectrum
             """),
-        ("cscale", 1, array(None, float), this,
-            """array of size (nseg, ndegree): Continumm polynomial coefficients for each wavelength segment
-            The x coordinates of each polynomial are chosen so that x = 0, at the first wavelength point,
-            i.e. x is shifted by wave[segment][0]
-            """),
-        ("cscale_flag", "none", lowercase(oneof(-3, -2, -1, 0, 1, 2, "none", "fix", "constant", "linear", "quadratic")), this,
+        ("vrad", 0, array(None, float), this, "array of size (nseg,): radial velocity of each segment in km/s"),
+        ("cscale_flag", "none", lowercase(oneof(-3, -2, -1, 0, 1, 2, 3, 4, 5, "none", "fix", "constant", "linear", "quadratic", "cubic", "quintic", "quantic")), this,
             """str: Flag that describes how to correct for the continuum
 
             allowed values are:
@@ -266,21 +159,26 @@ class SME_Structure(Parameters):
               * "whole": Fit the whole synthetic spectrum to the observation to determine the best fit
               * "mask": Fit a polynomial to the pixels marked as continuum in the mask
             """),
+        ("cscale", 1, array(None, float), this,
+            """array of size (nseg, ndegree): Continumm polynomial coefficients for each wavelength segment
+            The x coordinates of each polynomial are chosen so that x = 0, at the first wavelength point,
+            i.e. x is shifted by wave[segment][0]
+            """),
         ("normalize_by_continuum", True, asbool, this,
             "bool: Whether to normalize the synthetic spectrum by the synthetic continuum spectrum or not"),
         ("specific_intensities_only", False, asbool, this,
             "bool: Whether to keep the specific intensities or integrate them together"),
         ("gam6", 1, asfloat, this, "float: van der Waals scaling factor"),
         ("h2broad", True, asbool, this, "bool: Whether to use H2 broadening or not"),
-        ("accwi", 0.003, asfloat, this,
-            "float: minimum accuracy for linear spectrum interpolation vs. wavelength. Values below 1e-4 are not meaningful."),
-        ("accrt", 0.001, asfloat, this,
-            "float: minimum accuracy for synthethized spectrum at wavelength grid points in sme.wint. Values below 1e-4 are not meaningful."),
+        ("accwi", 3e-3, asfloat, this,
+            "float: minimum accuracy for linear spectrum interpolation vs. wavelength."),
+        ("accrt", 1e-4, asfloat, this,
+            "float: minimum accuracy for synthethized spectrum at wavelength grid points in sme.wint."),
         ("iptype", None, lowercase(oneof(None, "gauss", "sinc", "table")), this, "str: instrumental broadening type"),
-        ("ipres", 0, asfloat, this, "float: Instrumental resolution for instrumental broadening"),
+        ("ipres", 0, array(None, float), this, "float, array: Instrumental resolution for instrumental broadening"),
         ("ip_x", None, this, this, "array: Instrumental broadening table in x direction"),
         ("ip_y", None, this, this, "array: Instrumental broadening table in y direction"),
-        ("mu", np.geomspace(0.01, 1, 7), array(None, float), this,
+        ("mu", np.sqrt(0.5 * (2 * np.arange(7) + 1) / 7), array(None, float), this,
             """array of size (nmu,): Mu values to calculate radiative transfer at
             mu values describe the distance from the center of the stellar disk to the edge
             with mu = cos(theta), where theta is the angle of the observation,
@@ -294,6 +192,8 @@ class SME_Structure(Parameters):
             "Iliffe_vector of shape (nseg, ...): observed spectrum"),
         ("uncs", None, vector, this,
             "Iliffe_vector of shape (nseg, ...): uncertainties of the observed spectrum"),
+        ("telluric", None, vector, this, 
+            "Illife_vector of shape (nseg, ...): telluric spectrum that is multiplied with synth during the fit"),
         ("mask", None, vector, this,
             "Iliffe_vector of shape (nseg, ...): mask defining good and bad points for the fit"),
         ("synth", None, vector, this,
@@ -301,12 +201,72 @@ class SME_Structure(Parameters):
         ("cont", None, vector, this,
             "Iliffe_vector of shape (nseg, ...): continuum intensities"),
         ("linelist", LineList(), astype(LineList), this, "LineList: spectral line information"),
-        ("fitparameters", [], astype(list), this, "list: parameters to fit"),
+        ("fitparameters", [], astype(list, allow_None=True), this, "list: parameters to fit"),
+        ("fitresults", Fitresults(), astype(Fitresults), this, "Fitresults: fit results data"),
         ("atmo", Atmosphere(), astype(Atmosphere), this, "Atmosphere: model atmosphere data"),
         ("nlte", NLTE(), astype(NLTE), this, "NLTE: nlte calculation data"),
         ("system_info", Version(), astype(Version), this,
-            "Version: information about the host system running the calculation for debugging"),
-        ("citation_info", r"""
+            "Version: information about the host system running the calculation for debugging")
+    ]
+    # fmt: on
+
+    #:dict(str, int): Mask value specifier used in mob
+    mask_values = {"bad": 0, "line": 1, "continuum": 2}
+
+    def __init__(self, **kwargs):
+        wind = kwargs.get("wind", None)
+
+        atmo = kwargs.pop("atmo", {})
+        nlte = kwargs.pop("nlte", {})
+        idlver = kwargs.pop("idlver", {})
+        self.wave = None
+        self.wran = None
+        super().__init__(**kwargs)
+
+        if wind is not None and self.wave is not None:
+            wind = np.array([0, *(wind + 1)])
+            self.wave = Iliffe_vector(values=self.wave.ravel(), index=wind)
+
+        self.spec = kwargs.get("sob", None)
+        self.uncs = kwargs.get("uob", None)
+        self.mask = kwargs.get("mob", None)
+        self.synth = kwargs.get("smod", None)
+
+        self.meta["object"] = kwargs.get("obs_name", "")
+        try:
+            self.linelist = LineList(**kwargs)
+        except (KeyError, AttributeError):
+            # TODO ignore the warning during loading of data
+            logger.warning("No or incomplete linelist data present")
+
+        # Parse free parameters into one list
+        pname = kwargs.get("pname", [])
+        glob_free = kwargs.get("glob_free", [])
+        ab_free = kwargs.get("ab_free", [])
+        if len(ab_free) != 0:
+            ab_free = [f"abund {el}" for i, el in zip(ab_free, abund_elem) if i == 1]
+        fitparameters = np.concatenate((pname, glob_free, ab_free)).astype("U")
+        #:array of size (nfree): Names of the free parameters
+        self.fitparameters = np.unique(fitparameters)
+
+        self.fitresults = Fitresults(
+            maxiter=kwargs.get("maxiter", None),
+            chisq=kwargs.get("chisq", 0),
+            uncertainties=kwargs.get("punc", None),
+            covariance=kwargs.get("covar", None),
+        )
+
+        self.normalize_by_continuum = kwargs.get("cscale_flag", "") != "fix"
+
+        self.system_info = Version(**idlver)
+        atmo_abund = atmo.pop("abund", kwargs.get("abund", "empty"))
+        atmo_monh = atmo.pop(
+            "monh", atmo.pop("feh", kwargs.get("monh", kwargs.get("feh", 0)))
+        )
+        self.atmo = Atmosphere(**atmo, abund=atmo_abund, monh=atmo_monh,)
+        self.nlte = NLTE(**nlte)
+
+        self.citation_info = r"""
             @ARTICLE{2017A&A...597A..16P,
                 author = {{Piskunov}, Nikolai and {Valenti}, Jeff A.},
                 title = "{Spectroscopy Made Easy: Evolution}",
@@ -336,63 +296,7 @@ class SME_Structure(Parameters):
                 adsurl = {https://ui.adsabs.harvard.edu/abs/1996A&AS..118..595V},
                 adsnote = {Provided by the SAO/NASA Astrophysics Data System}
             }
-            """, asstr, this, "str: BibTex entry for SME")
-    ]
-    # fmt: on
-
-    #:dict(str, int): Mask value specifier used in mob
-    mask_values = {"bad": 0, "line": 1, "continuum": 2}
-
-    def __init__(self, **kwargs):
-        wind = kwargs.get("wind", None)
-
-        atmo = kwargs.pop("atmo", {})
-        nlte = kwargs.pop("nlte", {})
-        idlver = kwargs.pop("idlver", {})
-        super().__init__(**kwargs)
-
-        if wind is not None and self.wave is not None:
-            wind = np.array([0, *(wind + 1)])
-            self.wave = Iliffe_vector(values=self.wave.ravel(), index=wind)
-
-        self.spec = kwargs.get("sob", None)
-        self.uncs = kwargs.get("uob", None)
-        self.mask = kwargs.get("mob", None)
-        self.synth = kwargs.get("smod", None)
-
-        self.object = kwargs.get("obs_name", "")
-        try:
-            self.linelist = LineList(**kwargs)
-        except (KeyError, AttributeError):
-            # TODO ignore the warning during loading of data
-            logger.warning("No or incomplete linelist data present")
-
-        # Parse free parameters into one list
-        pname = kwargs.get("pname", [])
-        glob_free = kwargs.get("glob_free", [])
-        ab_free = kwargs.get("ab_free", [])
-        if len(ab_free) != 0:
-            ab_free = [f"abund {el}" for i, el in zip(ab_free, abund_elem) if i == 1]
-        fitparameters = np.concatenate((pname, glob_free, ab_free)).astype("U")
-        #:array of size (nfree): Names of the free parameters
-        self.fitparameters = np.unique(fitparameters)
-
-        self.fitresults = Fitresults(
-            maxiter=kwargs.get("maxiter", 0),
-            chisq=kwargs.get("chisq", 0),
-            uncertainties=kwargs.get("punc", None),
-            covar=kwargs.get("covar", None),
-        )
-
-        self.normalize_by_continuum = kwargs.get("cscale_flag", "") != "fix"
-
-        self.system_info = Version(**idlver)
-        self.atmo = Atmosphere(
-            **atmo,
-            abund=kwargs.get("abund", "empty"),
-            monh=kwargs.get("monh", kwargs.get("feh", 0)),
-        )
-        self.nlte = NLTE(**nlte)
+            """
 
         # Apply final conversions from IDL to Python version
         if "wave" in self:
@@ -438,7 +342,8 @@ class SME_Structure(Parameters):
                     values[i] = [self.wave[i][0], self.wave[i][-1]]
                 else:
                     values[i] = self.__wran[i]
-            return values
+            self.__wran = values
+            return self.__wran
         return self.__wran
 
     @_wran.setter
@@ -458,7 +363,8 @@ class SME_Structure(Parameters):
         nseg = self.nseg if self.nseg is not None else 1
 
         if self.__vrad is None:
-            return np.zeros(nseg)
+            self.__vrad = np.zeros(nseg)
+            return self.__vrad
 
         if self.vrad_flag == "none":
             return np.zeros(nseg)
@@ -470,7 +376,8 @@ class SME_Structure(Parameters):
             rv = np.zeros(self.nseg)
             rv[:nseg] = self.__vrad[:nseg]
             rv[nseg:] = self.__vrad[-1]
-            return rv
+            self.__vrad = rv
+            return self.__vrad
 
         return self.__vrad
 
@@ -480,6 +387,10 @@ class SME_Structure(Parameters):
 
     @property
     def _vrad_flag(self):
+        try:
+            _ = self.__vrad_flag
+        except AttributeError:
+            self.__vrad_flag = "none"
         return self.__vrad_flag
 
     @_vrad_flag.setter
@@ -499,6 +410,7 @@ class SME_Structure(Parameters):
         if self.__cscale is None:
             cs = np.zeros((nseg, self.cscale_degree + 1))
             cs[:, -1] = 1
+            self.__cscale = cs
             return cs
 
         if self.cscale_flag == "none":
@@ -519,9 +431,13 @@ class SME_Structure(Parameters):
         else:
             cs[:ns, :] = self.__cscale[:ns, :]
 
-        cs[ns:, -1] = self.__cscale[-1, -1]
+        cs[ns:, -1] = 1
 
-        return cs
+        # We need to update our internal representation as well
+        # since we might operate on that array
+        self.__cscale = cs
+
+        return self.__cscale
 
     @_cscale.setter
     def _cscale(self, value):
@@ -529,6 +445,10 @@ class SME_Structure(Parameters):
 
     @property
     def _cscale_flag(self):
+        try:
+            _ = self.__cscale_flag
+        except AttributeError:
+            self.__cscale_flag = "none"
         return self.__cscale_flag
 
     @_cscale_flag.setter
@@ -541,9 +461,12 @@ class SME_Structure(Parameters):
                 0: "constant",
                 1: "linear",
                 2: "quadratic",
+                3: "cubic",
+                4: "quintic",
+                5: "quantic",
             }[value]
-        if value == "quadratic":
-            logger.warning("Quadratic continuum scale is experimental")
+        if value in ["quadratic", "cubic", "quintic", "quantic"]:
+            logger.warning(f"{value} continuum scale is experimental")
 
         self.__cscale_flag = value
 
@@ -557,8 +480,24 @@ class SME_Structure(Parameters):
             raise ValueError("All values must be positive")
         if np.any(value > 1):
             raise ValueError("All values must be smaller or equal to 1")
+        # The mu values are expected in decreasing order
+        # For spherical models
+        value = np.sort(value)[::-1]
         self.__mu = value
 
+    @property
+    def _ipres(self):
+        return self.__ipres
+
+    @_ipres.setter
+    def _ipres(self, value):
+        size = np.size(value)
+        if self.nseg != 0 and size != 1 and size != self.nseg:
+            raise ValueError(
+                f"The instrument resolution must have 1 or {self.nseg} elements"
+            )
+        self.__ipres = value
+
     # Additional properties
     @property
     def nseg(self):
@@ -572,6 +511,10 @@ class SME_Structure(Parameters):
     def nmu(self):
         return self.mu.size
 
+    @nmu.setter
+    def nmu(self, value):
+        self.mu = np.sqrt(0.5 * (2 * np.arange(value) + 1) / value)
+
     @property
     def mask_good(self):
         if self.mask is None:
@@ -605,6 +548,12 @@ class SME_Structure(Parameters):
             return 1
         if self.cscale_flag == "quadratic":
             return 2
+        if self.cscale_flag == "cubic":
+            return 3
+        if self.cscale_flag == "quintic":
+            return 4
+        if self.cscale_flag == "quantic":
+            return 5
         if self.cscale_flag == "fix":
             # Use the underying element to avoid a loop
             if self.__cscale is not None:
